"Test convert_dict_items_to_datetime helper"
from typing import Any, Dict

import pytest

from metis_client.helpers import convert_dict_items_to_datetime, parse_rfc3339

testdata = [
    ({"some": 123}, {"some": 123}),
    (
        {"date": "2001-02-03 04:05:06", "str": "str", "int": int},
        {"date": parse_rfc3339("2001-02-03 04:05:06"), "str": "str", "int": int},
    ),
    (
        {"date": "2001-02-03 04:05:06", "other_date": "2001-02-03T04:05:06Z"},
        {
            "date": parse_rfc3339("2001-02-03 04:05:06"),
            "other_date": parse_rfc3339("2001-02-03T04:05:06Z"),
        },
    ),
    ({"date": ""}, {"date": None}),
    ({"date": None}, {"date": None}),
]


@pytest.mark.parametrize("x,expected", testdata)
def test_convert(x: Dict[str, Any], expected):
    "Test convert_dict_items_to_datetime()"
    y = convert_dict_items_to_datetime(["date", "other_date"], x)
    assert isinstance(y, type(x)), "Types should match"
    assert y == expected, "Converted should match expected value"



"Test MetisDatasourcesNamespace"

import asyncio
import json
from contextlib import nullcontext as does_not_raise
from copy import deepcopy
from datetime import datetime
from typing import List

import pytest
from aiohttp import web
from aiohttp.test_utils import TestClient, TestServer
from aiohttp.web_exceptions import HTTPNotFound, HTTPOk
from yarl import URL

from metis_client import MetisAPI, MetisAPIAsync, MetisTokenAuth
from metis_client.dtos import (
    MetisDataSourceContentOnlyDTO,
    MetisDataSourceDTO,
    MetisErrorDTO,
    MetisRequestIdDTO,
)
from metis_client.exc import (
    MetisNotFoundException,
    MetisPayloadException,
    MetisQuotaException,
)
from metis_client.models import MetisMessageEvent
from tests.helpers import random_word

dt = datetime.fromordinal(1)
DS_ID = 1
PARENT_DS_ID = 2
CHILD_DS_ID = 3
TOKEN = random_word(10)
PATH_STREAM = "/stream"
PATH_PING = "/v0"
PATH_DS = "/v0/datasources"
PATH_DS_ID = "/v0/datasources/{id}"
PATH_DS_POST_RESPONSE_PAYLOAD: MetisDataSourceDTO = {
    "id": DS_ID,
    "userId": 1,
    "userFirstName": random_word(10),
    "userLastName": random_word(10),
    "userEmail": random_word(10),
    "name": random_word(10),
    "content": random_word(10),
    "type": 1,
    "collections": [],
    "createdAt": dt,
    "updatedAt": dt,
    "parents": [PARENT_DS_ID],
    "children": [CHILD_DS_ID],
}
PATH_DS_POST_RESPONSE_ERROR_PAYLOAD: MetisErrorDTO = {"status": 400, "error": "oops"}
PATH_DS_GET_RESPONSE_PAYLOAD = deepcopy(PATH_DS_POST_RESPONSE_PAYLOAD)
PATH_DS_GET_RESPONSE_ERROR_PAYLOAD: MetisErrorDTO = {
    "status": 402,
    "error": {"message": "oops"},
}
PATH_DS_DELETE_RESPONSE_ERROR_PAYLOAD = deepcopy(PATH_DS_POST_RESPONSE_ERROR_PAYLOAD)
PATH_DS_DELETE_RESPONSE_ERROR404_PAYLOAD = {
    "status": HTTPNotFound.status_code,
    "error": "not found",
}
PATH_DS_ID_GET_RESPONSE: MetisDataSourceContentOnlyDTO = {"content": random_word(10)}

event_stream: List[MetisMessageEvent] = []


def make_datasources_event(req_id: str, datas) -> MetisMessageEvent:
    "Create datasources event with content"
    evt_data_dto = {"reqId": req_id, "data": datas, "total": 1, "types": []}
    return MetisMessageEvent(
        "datasources", "datasources", json.dumps(evt_data_dto), "", ""
    )


def make_error_event(req_id: str, datas) -> MetisMessageEvent:
    "Create errors event with content"
    evt_data_dto = {"reqId": req_id, "data": datas}
    return MetisMessageEvent("errors", "errors", json.dumps(evt_data_dto), "", "")


async def sse_handler(_: web.Request) -> web.Response:
    "Request handler"
    body = "".join(
        [f"event: {evt.type}\ndata: {evt.data}\n\n\n" for evt in event_stream]
    )
    resp = web.Response(status=HTTPOk.status_code, body=body)
    resp.content_type = "text/event-stream"
    resp.headers["Cache-Control"] = "no-store"
    return resp


async def ping_handler(_) -> web.Response:
    "Request handler"
    evt = MetisMessageEvent("", "ping", "", "", "")
    event_stream.append(evt)
    return web.Response(status=HTTPOk.status_code)


async def datasource_create_handler(request: web.Request) -> web.Response:
    "Request handler"
    body: MetisRequestIdDTO = {"reqId": random_word(10)}

    payload = await request.json()
    if payload.get("content") == "ok":
        ds_dto = {
            **PATH_DS_POST_RESPONSE_PAYLOAD,
            "createdAt": dt.isoformat(),
            "updatedAt": dt.isoformat(),
        }
        evt = make_datasources_event(body["reqId"], [ds_dto])
    else:
        evt = make_error_event(body["reqId"], [PATH_DS_POST_RESPONSE_ERROR_PAYLOAD])

    event_stream.append(evt)

    return web.json_response(body, status=HTTPOk.status_code)


async def list_datasources_handler(request: web.Request) -> web.Response:
    "Request handler"
    auth_header = request.headers.get("Authorization", "")
    body: MetisRequestIdDTO = {"reqId": random_word(10)}

    if auth_header != f"Bearer {str(True)}":
        ds_dto = {
            **PATH_DS_GET_RESPONSE_PAYLOAD,
            "createdAt": dt.isoformat(),
            "updatedAt": dt.isoformat(),
        }
        evt = make_datasources_event(body["reqId"], [ds_dto])
    else:
        evt = make_error_event(body["reqId"], [PATH_DS_GET_RESPONSE_ERROR_PAYLOAD])

    event_stream.append(evt)

    return web.json_response(body, status=HTTPOk.status_code)


async def delete_datasource_handler(request: web.Request) -> web.Response:
    "Request handler"

    if request.match_info.get("id") not in ["1", "2"]:
        return web.Response(status=HTTPNotFound.status_code)

    body: MetisRequestIdDTO = {"reqId": random_word(10)}

    if request.match_info.get("id") == "2":
        evt = make_error_event(body["reqId"], [PATH_DS_DELETE_RESPONSE_ERROR_PAYLOAD])
    else:
        evt = make_datasources_event(body["reqId"], [])

    event_stream.append(evt)

    return web.json_response(body, status=HTTPOk.status_code)


async def get_datasource_handler(request: web.Request) -> web.Response:
    "Request handler"
    if request.match_info.get("id") == "2":
        # stupid variant of success with error
        return web.json_response(
            PATH_DS_DELETE_RESPONSE_ERROR404_PAYLOAD, status=HTTPOk.status_code
        )
    if request.match_info.get("id") not in ["1", "2"]:
        return web.json_response(
            PATH_DS_DELETE_RESPONSE_ERROR404_PAYLOAD, status=HTTPNotFound.status_code
        )
    return web.json_response(PATH_DS_ID_GET_RESPONSE, status=HTTPOk.status_code)


async def create_app() -> web.Application:
    "Create web application"
    app = web.Application()
    app.router.add_head(PATH_PING, ping_handler)
    app.router.add_get(PATH_STREAM, sse_handler)
    app.router.add_post(PATH_DS, datasource_create_handler)
    app.router.add_get(PATH_DS, list_datasources_handler)
    app.router.add_get(PATH_DS_ID, get_datasource_handler)
    app.router.add_delete(PATH_DS_ID, delete_datasource_handler)
    return app


@pytest.fixture
async def aiohttp_client_impl(aiohttp_client) -> TestClient:
    "Create test client"
    app = await create_app()
    return await aiohttp_client(TestServer(app))


@pytest.fixture
def base_url(aiohttp_client_impl: TestClient) -> URL:
    "Return base url"
    return aiohttp_client_impl.make_url("")


@pytest.fixture
def client(base_url: URL) -> MetisAPI:
    "Return sync client"
    return MetisAPI(base_url, auth=MetisTokenAuth(TOKEN))


@pytest.fixture
async def client_async(base_url: URL):
    "Return sync client"
    async with MetisAPIAsync(base_url, auth=MetisTokenAuth(TOKEN)) as client:
        yield client


@pytest.mark.parametrize(
    "content, expected, raises",
    [
        ("ok", PATH_DS_POST_RESPONSE_PAYLOAD, does_not_raise()),
        ("fail", None, pytest.raises(MetisPayloadException)),
    ],
)
async def test_create_datasource(
    client: MetisAPI, client_async: MetisAPIAsync, content, expected, raises
):
    "Test create()"
    with raises:
        src = await client_async.v0.datasources.create(content)
        assert src == expected, "Response matches"
    with raises:
        src = await asyncio.get_event_loop().run_in_executor(
            None, client.v0.datasources.create, content
        )
        assert src == expected, "Response matches"


@pytest.mark.parametrize(
    "fail, expected, raises",
    [
        (False, [PATH_DS_GET_RESPONSE_PAYLOAD], does_not_raise()),
        (True, None, pytest.raises(MetisQuotaException)),
    ],
)
async def test_list_datasources(base_url: URL, fail: bool, expected, raises):
    "Test list()"
    with raises:
        async with MetisAPIAsync(base_url, auth=MetisTokenAuth(str(fail))) as client:
            src = await client.v0.datasources.list()
        assert src == expected, "Response matches"
    with raises:
        client = MetisAPI(base_url, auth=MetisTokenAuth(str(fail)))
        src = await asyncio.get_event_loop().run_in_executor(
            None, client.v0.datasources.list
        )
        assert src == expected, "Response matches"


@pytest.mark.parametrize(
    "method_name, data_id, fail, expected, raises",
    [
        # get()
        ("get", -1, True, None, pytest.raises(MetisQuotaException)),
        ("get", DS_ID, False, PATH_DS_GET_RESPONSE_PAYLOAD, does_not_raise()),
        ("get", PARENT_DS_ID, False, None, does_not_raise()),
        # get_parents()
        (
            "get_parents",
            CHILD_DS_ID,
            False,
            [PATH_DS_GET_RESPONSE_PAYLOAD],
            does_not_raise(),
        ),
        ("get_parents", DS_ID, False, [], does_not_raise()),
        # get_children()
        (
            "get_children",
            PARENT_DS_ID,
            False,
            [PATH_DS_GET_RESPONSE_PAYLOAD],
            does_not_raise(),
        ),
        ("get_children", DS_ID, False, [], does_not_raise()),
    ],
)
async def test_get_datasource(
    base_url: URL, method_name: str, data_id: int, fail: bool, expected, raises
):
    "Test get(), get_parents(), get_children()"
    with raises:
        async with MetisAPIAsync(base_url, auth=MetisTokenAuth(str(fail))) as client:
            method = getattr(client.v0.datasources, method_name)
            src = await method(data_id)
        assert src == expected, "Response matches"
    with raises:
        client = MetisAPI(base_url, auth=MetisTokenAuth(str(fail)))
        method = getattr(client.v0.datasources, method_name)
        src = await asyncio.get_event_loop().run_in_executor(None, method, data_id)
        assert src == expected, "Response matches"


@pytest.mark.parametrize(
    "ds_id, raises",
    [
        (1, does_not_raise()),
        (2, pytest.raises(MetisPayloadException)),
        (3, pytest.raises(MetisNotFoundException)),
    ],
)
async def test_delete_datasource(
    client: MetisAPI, client_async: MetisAPIAsync, ds_id: int, raises
):
    "Test delete()"
    with raises:
        await client_async.v0.datasources.delete(ds_id)
    with raises:
        await asyncio.get_event_loop().run_in_executor(
            None, client.v0.datasources.delete, ds_id
        )


@pytest.mark.parametrize(
    "ds_id, raises",
    [
        (1, does_not_raise()),
        (2, pytest.raises(MetisNotFoundException)),
        (3, pytest.raises(MetisNotFoundException)),
    ],
)
async def test_get_datasource_contents(
    client: MetisAPI, client_async: MetisAPIAsync, ds_id, raises
):
    "Test get_content()"
    with raises:
        res = await client_async.v0.datasources.get_content(ds_id)
        assert res == PATH_DS_ID_GET_RESPONSE, "Response matches"
    with raises:
        res = await asyncio.get_event_loop().run_in_executor(
            None, client.v0.datasources.get_content, ds_id
        )
        assert res == PATH_DS_ID_GET_RESPONSE, "Response matches"



#!/usr/bin/env python3
"""Example of usage of asynchronous client"""

import asyncio
import sys

from metis_client import MetisAPIAsync, MetisTokenAuth
from metis_client.dtos.calculation import MetisCalculationDTO

API_URL = "http://localhost:3000"
TEST_ENGINE = "dummy"

try:
    with open(sys.argv[1], encoding="utf-8") as fp:
        CONTENT = fp.read()
except IndexError:
    # pylint: disable=line-too-long
    CONTENT = """{"attributes":{"immutable_id":42, "species":[{"chemical_symbols":
["Au"]}],"cartesian_site_positions":[[0,0,0]],"lattice_vectors":[[0,2,2],[2,0,2],[2,2,0]]}}"""


async def on_progress_log(calc: MetisCalculationDTO):
    "Print progress"
    print("Progress:", calc.get("progress"))


async def create_calc_then_get_results(client: MetisAPIAsync):
    "Create data source, run calculation, then wait for the results"

    data = await client.v0.datasources.create(CONTENT)
    assert data

    calc = await client.v0.calculations.create(data.get("id"), engine=TEST_ENGINE)
    assert calc

    results = await client.v0.calculations.get_results(
        calc["id"], on_progress=on_progress_log
    )
    print(results)
    assert results
    print("=" * 50 + "Test passed")


async def create_calc_and_get_results(client: MetisAPIAsync):
    "Create data source, run calculation and wait for the results"

    data = await client.v0.datasources.create(CONTENT)
    assert data

    results = await client.v0.calculations.create_get_results(
        data["id"], engine=TEST_ENGINE, on_progress=on_progress_log
    )
    print(results)
    assert results
    print("=" * 50 + "Test passed")


async def create_calc_then_cancel(client: MetisAPIAsync):
    """
    Create data source. Run calculation.
    Stop watching calculation on 50%, cancel calculation.
    """

    data = await client.v0.datasources.create(CONTENT)
    assert data

    async def stop_watching_on_half(calc: MetisCalculationDTO):
        await on_progress_log(calc)
        return calc.get("progress") < 50

    calc = await client.v0.calculations.create(data.get("id"), engine=TEST_ENGINE)
    assert calc

    results = await client.v0.calculations.get_results(
        calc["id"], on_progress=stop_watching_on_half
    )
    assert results is None
    await client.v0.calculations.cancel(calc["id"])
    print("=" * 50 + "Test passed")


async def main():
    "Run all examples"
    async with MetisAPIAsync(API_URL, auth=MetisTokenAuth("admin@test.com")) as client:
        print(await client.v0.auth.whoami())
        print(
            "The following engines are available:",
            await client.v0.calculations.get_engines(),
        )

        await create_calc_then_get_results(client)
        await create_calc_and_get_results(client)
        await create_calc_then_cancel(client)


asyncio.run(main())



"Test MetisLocalUserAuth"
from itertools import count
from json import JSONDecodeError

import pytest
from aiohttp import CookieJar, web
from aiohttp.test_utils import TestClient, TestServer
from aiohttp.web_exceptions import (
    HTTPBadRequest,
    HTTPOk,
    HTTPTooManyRequests,
    HTTPUnauthorized,
)
from freezegun import freeze_time

from metis_client import MetisLocalUserAuth
from tests.helpers import random_word

EMAIL = random_word(10)
PASSWORD = random_word(10)

AUTH_PATH = f"/{MetisLocalUserAuth._endpoint}"  # pylint: disable=protected-access

TOO_MANY_COUNTER = count(0)


async def handler(request: web.Request) -> web.Response:
    "Request handler"
    # bad request
    try:
        payload = await request.json()
        if "email" not in payload or "password" not in payload:
            raise TypeError
    except (JSONDecodeError, TypeError):
        resp = web.Response(
            body=HTTPBadRequest.status, status=HTTPBadRequest.status_code
        )
        resp.set_cookie("_sid", "session but bad")
        return resp

    # simulate too many requests on the first request
    if next(TOO_MANY_COUNTER) < 1:
        resp = web.Response(body="not ok", status=HTTPTooManyRequests.status_code)
        resp.set_cookie("_sid", "session but bad")
        return resp

    # bad credentials
    if payload.get("email") != EMAIL or payload.get("password") != PASSWORD:
        resp = web.Response(body="not ok", status=HTTPUnauthorized.status_code)
        resp.set_cookie("_sid", "session but bad")
        return resp

    # ok
    resp = web.Response(body="ok", status=HTTPOk.status_code)
    resp.set_cookie("_sid", "ok")
    return resp


async def create_app() -> web.Application:
    "Create web application"
    app = web.Application()
    app.router.add_post(AUTH_PATH, handler)
    return app


@pytest.fixture
async def cli(aiohttp_client) -> TestClient:
    "Create test client"
    jar = CookieJar(unsafe=True, treat_as_secure_origin="http://127.0.0.1")
    return await aiohttp_client(TestServer(await create_app()), cookie_jar=jar)


@freeze_time("1970-01-01", auto_tick_seconds=10)
async def test_auth_ok(cli: TestClient):
    "Test successful authentication"
    authenticator = MetisLocalUserAuth(EMAIL, PASSWORD)
    base_url = cli.make_url("")
    assert await authenticator.should_update(
        cli.session, base_url
    ), "Update is needed before authentication"
    assert await authenticator.authenticate(
        cli.session, base_url
    ), "Correct email-password pair should authenticate"
    assert (
        await authenticator.should_update(cli.session, base_url) is False
    ), "Update is not needed after authentication"


async def test_auth_not_ok(cli: TestClient):
    "Test unsuccessful authentication"
    authenticator = MetisLocalUserAuth(EMAIL, "")
    base_url = cli.make_url("")
    assert (
        await authenticator.authenticate(cli.session, base_url) is False
    ), "Incorrect email-password pair should fail to authenticate"
    assert await authenticator.should_update(
        cli.session, base_url
    ), "Update is needed after failed authentication"



"Test MetisTokenAuth"

import pytest
from aiohttp import web
from aiohttp.test_utils import TestClient, TestServer

from metis_client import MetisTokenAuth
from tests.helpers import random_word

TOKEN = random_word(10)


@pytest.fixture
async def cli(aiohttp_client) -> TestClient:
    "Create test client"
    return await aiohttp_client(TestServer(web.Application()))


async def test_auth(cli: TestClient):
    "Test successful authentication"
    authenticator = MetisTokenAuth(TOKEN)
    base_url = cli.make_url("")
    assert await authenticator.should_update(
        cli.session, base_url
    ), "Update is needed before authentication"
    assert (
        cli.session.headers.get("Authorization") is None
    ), "There is no Authorization header before authenticate"
    assert await authenticator.authenticate(
        cli.session, base_url
    ), "Token authentication is sucessful"
    assert (
        cli.session.headers.get("Authorization") == f"Bearer {TOKEN}"
    ), "There is the Authorization header with Bearer after authenticate"
    assert await authenticator.should_update(
        cli.session, base_url
    ), "Update is needed after authentication"



"Test MetisAuthNamespace"

import asyncio
from copy import deepcopy
from datetime import datetime

import pytest
from aiohttp import web
from aiohttp.test_utils import TestClient, TestServer
from aiohttp.web_exceptions import HTTPOk
from yarl import URL

from metis_client import MetisAPI, MetisAPIAsync, MetisTokenAuth
from metis_client.dtos import MetisUserDTO
from metis_client.dtos.auth import MetisAuthCredentialsRequestDTO
from metis_client.exc import MetisAuthenticationException
from tests.helpers import random_word

dt = datetime.fromordinal(1)
TOKEN = random_word(10)
PATH_AUTH = "/v0/auth"
PATH_AUTH_GET_PAYLOAD: MetisUserDTO = {
    "id": 1,
    "firstName": random_word(10),
    "lastName": random_word(10),
    "email": random_word(10),
    "emailVerified": True,
    "roleLabel": random_word(10),
    "roleSlug": random_word(10),
    "permissions": {},
    "provider": "",
    "createdAt": dt,
    "updatedAt": dt,
}
PATH_AUTH_POST_PAYLOAD: MetisAuthCredentialsRequestDTO = {
    "email": random_word(10),
    "password": random_word(10),
}


async def auth_whoami_handler(_: web.Request) -> web.Response:
    "Request handler"
    payload = {
        **deepcopy(PATH_AUTH_GET_PAYLOAD),
        "createdAt": PATH_AUTH_GET_PAYLOAD.get("createdAt", dt).isoformat(),
        "updatedAt": PATH_AUTH_GET_PAYLOAD.get("updatedAt", dt).isoformat(),
    }
    return web.json_response(payload, status=HTTPOk.status_code)


async def auth_login_handler(request: web.Request) -> web.Response:
    "Request handler"
    payload = await request.json()
    if payload == PATH_AUTH_POST_PAYLOAD:
        return web.Response(status=HTTPOk.status_code)
    return web.Response(status=web.HTTPUnauthorized.status_code)


async def create_app() -> web.Application:
    "Create web application"
    app = web.Application()
    app.router.add_get(PATH_AUTH, auth_whoami_handler)
    app.router.add_post(PATH_AUTH, auth_login_handler)
    return app


@pytest.fixture
async def aiohttp_client(aiohttp_client) -> TestClient:
    "Create test client"
    app = await create_app()
    return await aiohttp_client(TestServer(app))


@pytest.fixture
def base_url(aiohttp_client: TestClient) -> URL:
    "Return base url"
    return aiohttp_client.make_url("")


@pytest.fixture
def client(base_url: URL) -> MetisAPI:
    "Return sync client"
    return MetisAPI(base_url, auth=MetisTokenAuth(TOKEN))


@pytest.fixture
async def client_async(base_url: URL):
    "Return sync client"
    async with MetisAPIAsync(base_url, auth=MetisTokenAuth(TOKEN)) as client:
        yield client


async def test_whoami(client: MetisAPI, client_async: MetisAPIAsync):
    "Test whoami()"
    whoami_async = await client_async.v0.auth.whoami()
    whoami = await asyncio.get_event_loop().run_in_executor(None, client.v0.auth.whoami)
    assert whoami_async == whoami == PATH_AUTH_GET_PAYLOAD, "Response matches"


async def test_login_ok(client: MetisAPI, client_async: MetisAPIAsync):
    "Test login()"
    args = (PATH_AUTH_POST_PAYLOAD["email"], PATH_AUTH_POST_PAYLOAD["password"])
    auth_async = await client_async.v0.auth.login(*args)
    auth_sync = await asyncio.get_event_loop().run_in_executor(
        None, client.v0.auth.login, *args
    )
    assert auth_async is True, "Auth should be successful"
    assert auth_async == auth_sync, "Results should be the same"


async def test_login_fail(client: MetisAPI, client_async: MetisAPIAsync):
    "Test login()"
    args = (random_word(10), random_word(10))
    with pytest.raises(MetisAuthenticationException):
        await client_async.v0.auth.login(*args)
    with pytest.raises(MetisAuthenticationException):
        await asyncio.get_event_loop().run_in_executor(
            None, client.v0.auth.login, *args
        )



"Test MetisHub"

from metis_client.models import MetisHub, MetisSubscription


async def test_close_hub():
    "Test close()"
    hub = MetisHub()
    hub.subscribe(MetisSubscription(hub))
    await hub.close()



"Test MetisCalculationsNamespace"

import asyncio
import json
from contextlib import nullcontext as does_not_raise
from copy import deepcopy
from datetime import datetime
from typing import List

import pytest
from aiohttp import web
from aiohttp.test_utils import TestClient, TestServer
from aiohttp.web_exceptions import HTTPNotFound, HTTPOk
from yarl import URL

from metis_client import MetisAPI, MetisAPIAsync, MetisTokenAuth
from metis_client.dtos import MetisCalculationDTO, MetisErrorDTO, MetisRequestIdDTO
from metis_client.dtos.datasource import DataSourceType
from metis_client.exc import MetisPayloadException, MetisQuotaException
from metis_client.models import MetisMessageEvent
from tests.helpers import random_word
from tests.namespaces.test_datasources import (
    DS_ID,
    PATH_DS_POST_RESPONSE_PAYLOAD,
    make_datasources_event,
)

dt = datetime.fromordinal(1)
TOKEN = random_word(10)
CALC_ID = 10
PATH_STREAM = "/stream"
PATH_PING = "/v0"
PATH_C = "/v0/calculations"
PATH_C_ID = "/v0/calculations/{id}"
PATH_C_ENGINES = "/v0/calculations/engines"
PATH_C_POST_RESPONSE_PAYLOAD: MetisCalculationDTO = {
    "id": CALC_ID,
    "name": random_word(10),
    "userId": 1,
    "progress": 1,
    "result": [],
    "createdAt": dt,
    "updatedAt": dt,
    "parent": DS_ID,
}
PATH_C_POST_RESPONSE_ERROR_PAYLOAD: MetisErrorDTO = {"status": 400, "error": "oops"}
PATH_C_GET_RESPONSE_PAYLOAD = deepcopy(PATH_C_POST_RESPONSE_PAYLOAD)
PATH_C_GET_RESPONSE_ERROR_PAYLOAD: MetisErrorDTO = {"status": 402, "error": "oops"}
PATH_C_DELETE_RESPONSE_ERROR_PAYLOAD = deepcopy(PATH_C_POST_RESPONSE_ERROR_PAYLOAD)
PATH_C_DELETE_RESPONSE_ERROR404_PAYLOAD = {
    "status": HTTPNotFound.status_code,
    "error": "not found",
}
PATH_C_GET_ENGINES_RESPONSE = ["dummy", "other"]

event_stream: List[MetisMessageEvent] = []


def make_calculations_event(req_id: str, datas) -> MetisMessageEvent:
    "Create calculations event with content"
    evt_data_dto = {"reqId": req_id, "data": datas, "total": 1, "types": []}
    return MetisMessageEvent(
        "calculations", "calculations", json.dumps(evt_data_dto), "", ""
    )


def make_error_event(req_id: str, datas) -> MetisMessageEvent:
    "Create errors event with content"
    evt_data_dto = {"reqId": req_id, "data": datas}
    return MetisMessageEvent("errors", "errors", json.dumps(evt_data_dto), "", "")


async def sse_handler(_: web.Request) -> web.Response:
    "Request handler"
    body = "".join(
        [f"event: {evt.type}\ndata: {evt.data}\n\n\n" for evt in event_stream]
    )
    resp = web.Response(status=HTTPOk.status_code, body=body)
    resp.content_type = "text/event-stream"
    resp.headers["Cache-Control"] = "no-store"
    return resp


async def ping_handler(_) -> web.Response:
    "Request handler"
    evt = MetisMessageEvent("", "", "pong", "", "")
    event_stream.append(evt)
    return web.Response(status=HTTPOk.status_code)


async def calculation_create_handler(request: web.Request) -> web.Response:
    "Request handler"
    body: MetisRequestIdDTO = {"reqId": random_word(10)}

    payload = await request.json()
    if payload.get("engine") == "dummy":
        cal_dto = {
            **PATH_C_POST_RESPONSE_PAYLOAD,
            "createdAt": dt.isoformat(),
            "updatedAt": dt.isoformat(),
        }
        evt = make_calculations_event(body["reqId"], [cal_dto])
        event_stream.append(evt)
    elif payload.get("engine") == "results":
        cal_dto = {
            **PATH_C_POST_RESPONSE_PAYLOAD,
            "createdAt": dt.isoformat(),
            "updatedAt": dt.isoformat(),
            "progress": 25,
            "parent": payload["dataId"],
        }
        event_stream.append(make_calculations_event(body["reqId"], [cal_dto]))
        event_stream.append(
            make_calculations_event(body["reqId"], [{**cal_dto, "progress": 50}])
        )
        event_stream.append(make_calculations_event(body["reqId"], []))
        event_stream.append(
            make_calculations_event(body["reqId"], [{**cal_dto, "progress": 75}])
        )
        ds_dto = {
            **PATH_DS_POST_RESPONSE_PAYLOAD,
            "createdAt": dt.isoformat(),
            "updatedAt": dt.isoformat(),
            "parents": [payload["dataId"]],
            "type": DataSourceType.PROPERTY,
        }
        event_stream.append(
            make_calculations_event(body["reqId"], [{**ds_dto, "progress": 100}])
        )
        event_stream.append(make_datasources_event(body["reqId"], [ds_dto]))
    else:
        evt = make_error_event(body["reqId"], [PATH_C_POST_RESPONSE_ERROR_PAYLOAD])
        event_stream.append(evt)

    return web.json_response(body, status=HTTPOk.status_code)


async def list_calculations_handler(request: web.Request) -> web.Response:
    "Request handler"
    auth_header = request.headers.get("Authorization", "")
    body: MetisRequestIdDTO = {"reqId": random_word(10)}

    if auth_header != f"Bearer {str(True)}":
        ds_dto = {
            **PATH_C_GET_RESPONSE_PAYLOAD,
            "createdAt": dt.isoformat(),
            "updatedAt": dt.isoformat(),
        }
        evt = make_calculations_event(body["reqId"], [ds_dto])
    else:
        evt = make_error_event(body["reqId"], [PATH_C_GET_RESPONSE_ERROR_PAYLOAD])

    event_stream.append(evt)

    return web.json_response(body, status=HTTPOk.status_code)


async def cancel_calculations_handler(request: web.Request) -> web.Response:
    "Request handler"

    if request.match_info.get("id") not in ["1", "2"]:
        return web.Response(status=HTTPNotFound.status_code)

    body: MetisRequestIdDTO = {"reqId": random_word(10)}

    if request.match_info.get("id") == "2":
        evt = make_error_event(body["reqId"], [PATH_C_DELETE_RESPONSE_ERROR_PAYLOAD])
    else:
        evt = make_calculations_event(body["reqId"], [])

    event_stream.append(evt)

    return web.json_response(body, status=HTTPOk.status_code)


async def get_engines_handler(_: web.Request) -> web.Response:
    "Request handler"
    return web.json_response(PATH_C_GET_ENGINES_RESPONSE, status=HTTPOk.status_code)


async def create_app() -> web.Application:
    "Create web application"
    app = web.Application()
    app.router.add_head(PATH_PING, ping_handler)
    app.router.add_get(PATH_STREAM, sse_handler)
    app.router.add_post(PATH_C, calculation_create_handler)
    app.router.add_get(PATH_C, list_calculations_handler)
    app.router.add_get(PATH_C_ENGINES, get_engines_handler)
    app.router.add_delete(PATH_C_ID, cancel_calculations_handler)
    return app


@pytest.fixture
async def aiohttp_client_impl(aiohttp_client) -> TestClient:
    "Create test client"
    app = await create_app()
    return await aiohttp_client(TestServer(app))


@pytest.fixture
def base_url(aiohttp_client_impl: TestClient) -> URL:
    "Return base url"
    return aiohttp_client_impl.make_url("")


@pytest.fixture
def client(base_url: URL) -> MetisAPI:
    "Return sync client"
    return MetisAPI(base_url, auth=MetisTokenAuth(TOKEN))


@pytest.fixture
async def client_async(base_url: URL):
    "Return sync client"
    async with MetisAPIAsync(base_url, auth=MetisTokenAuth(TOKEN)) as client:
        yield client


@pytest.mark.parametrize(
    "engine, expected, raises",
    [
        ("dummy", PATH_C_POST_RESPONSE_PAYLOAD, does_not_raise()),
        ("fail", None, pytest.raises(MetisPayloadException)),
    ],
)
async def test_create_calculation(
    client: MetisAPI, client_async: MetisAPIAsync, engine, expected, raises
):
    "Test create()"
    with raises:
        cal = await client_async.v0.calculations.create(1, engine)
        assert cal == expected, "Response matches"
    with raises:
        cal = await asyncio.get_event_loop().run_in_executor(
            None, client.v0.calculations.create, 1, engine
        )
        assert cal == expected, "Response matches"


@pytest.mark.parametrize(
    "early_exit, sync, create",
    [
        (True, True, True),
        (False, True, False),
        (False, False, True),
        (False, False, False),
    ],
)
async def test_get_results(
    client: MetisAPI,
    client_async: MetisAPIAsync,
    early_exit: bool,
    sync: bool,
    create: bool,
):
    "Test get_results() and create_get_results()"

    def on_progress(calc: MetisCalculationDTO):
        assert on_progress.progress <= calc.get("progress", 0)
        on_progress.progress = calc.get("progress", 0)
        return not early_exit

    async def async_on_progress(calc: MetisCalculationDTO):
        return on_progress(calc)

    on_progress.progress = 0

    if sync:
        if create:
            results = await asyncio.get_event_loop().run_in_executor(
                None,
                client.v0.calculations.create_get_results,
                DS_ID,
                "results",
                None,
                on_progress,
            )
        else:
            calc = await asyncio.get_event_loop().run_in_executor(
                None, client.v0.calculations.create, DS_ID, "results"
            )
            assert calc and calc["id"] == CALC_ID
            results = await asyncio.get_event_loop().run_in_executor(
                None, client.v0.calculations.get_results, calc["id"], on_progress
            )
    else:
        if create:
            results = await client_async.v0.calculations.create_get_results(
                DS_ID, "results", None, async_on_progress
            )
        else:
            calc = await client_async.v0.calculations.create(DS_ID, "results")
            assert calc and calc["id"] == CALC_ID
            results = await client_async.v0.calculations.get_results(
                calc["id"], async_on_progress
            )

    if early_exit:
        assert results is None
    else:
        assert results and DS_ID in results[0].get("parents", [])


@pytest.mark.parametrize(
    "fail, expected, raises",
    [
        (False, [PATH_C_GET_RESPONSE_PAYLOAD], does_not_raise()),
        (True, None, pytest.raises(MetisQuotaException)),
    ],
)
async def test_list_calculations(base_url: URL, fail: bool, expected, raises):
    "Test list()"
    with raises:
        async with MetisAPIAsync(base_url, auth=MetisTokenAuth(str(fail))) as client:
            cal = await client.v0.calculations.list()
        assert cal == expected, "Response matches"
    with raises:
        client = MetisAPI(base_url, auth=MetisTokenAuth(str(fail)))
        cal = await asyncio.get_event_loop().run_in_executor(
            None, client.v0.calculations.list
        )
        assert cal == expected, "Response matches"


@pytest.mark.parametrize(
    "calc_id, fail, expected, raises",
    [
        (CALC_ID, False, PATH_C_GET_RESPONSE_PAYLOAD, does_not_raise()),
        (-1, False, None, does_not_raise()),
        (-1, True, None, pytest.raises(MetisQuotaException)),
    ],
)
async def test_get_calculation(
    base_url: URL, calc_id: int, fail: bool, expected, raises
):
    "Test get()"
    with raises:
        async with MetisAPIAsync(base_url, auth=MetisTokenAuth(str(fail))) as client:
            cal = await client.v0.calculations.get(calc_id)
        assert cal == expected, "Response matches"
    with raises:
        client = MetisAPI(base_url, auth=MetisTokenAuth(str(fail)))
        cal = await asyncio.get_event_loop().run_in_executor(
            None, client.v0.calculations.get, calc_id
        )
        assert cal == expected, "Response matches"


@pytest.mark.parametrize(
    "cal_id, raises",
    [
        (1, does_not_raise()),
        (2, pytest.raises(MetisPayloadException)),
    ],
)
async def test_cancel_calculation(
    client: MetisAPI, client_async: MetisAPIAsync, cal_id: int, raises
):
    "Test delete()"
    with raises:
        await client_async.v0.calculations.cancel(cal_id)
    with raises:
        await asyncio.get_event_loop().run_in_executor(
            None, client.v0.calculations.cancel, cal_id
        )


async def test_get_engines(client: MetisAPI, client_async: MetisAPIAsync):
    "Test get_engines()"
    en_async = await client_async.v0.calculations.get_engines()
    en_sync = await asyncio.get_event_loop().run_in_executor(
        None, client.v0.calculations.get_engines
    )
    assert en_async == en_sync, "Response matches"
    assert en_sync == PATH_C_GET_ENGINES_RESPONSE, "Response matches"



"""Test MetisMessageEvent"""

# from metis_client.dtos import MetisMessageEvent
from metis_client.models import MetisMessageEvent


def test_encoding_error():
    "Test encoding error in to_dto()"

    evt = MetisMessageEvent("errors", "message", "data", "origin", "last_event_id")
    result = evt.to_dto()

    assert (
        isinstance(result["data"], dict)
        and result["data"].get("data", [{}])[-1].get("status") == 400
    ), "Shoult return error event with status 400"



"Test MetisAPIAsync"

import pytest

from metis_client import MetisAPIAsync, MetisNoAuth


async def test_relative_url():
    "Test relative url"
    with pytest.raises(TypeError):
        MetisAPIAsync("/relative", auth=MetisNoAuth())



"""Calculations endpoints namespace"""
import sys
from datetime import datetime
from functools import partial
from inspect import iscoroutinefunction
from typing import Awaitable, Callable, Optional, Union, cast

from ..dtos import (
    DataSourceType,
    MetisCalculationDTO,
    MetisDataSourceDTO,
    MetisRequestIdDTO,
)
from ..helpers import raise_on_metis_error
from ..models import act_and_get_result_from_stream
from .base import BaseNamespace

if sys.version_info < (3, 9):  # pragma: no cover
    from typing import Sequence
else:  # pragma: no cover
    from collections.abc import Sequence


DATA_SOURCE_CALC_RESULT_TYPES = [DataSourceType.PROPERTY, DataSourceType.PATTERN]

MetisCalculationOnProgressT = Callable[
    [MetisCalculationDTO], Union[Optional[bool], Awaitable[Optional[bool]]]
]


class MetisCalculationsNamespace(BaseNamespace):
    """Calculations endpoints namespace"""

    async def cancel_event(self, calc_id: int) -> MetisRequestIdDTO:
        "Cancel calculation"
        async with await self._client.request(
            method="DELETE",
            url=self._base_url / str(calc_id),
            auth_required=True,
        ) as resp:
            return await resp.json()

    async def cancel(self, calc_id: int) -> None:
        "Cancel calculation and wait for result"
        await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.cancel_event, calc_id)
        )

    async def create_event(
        self,
        data_id: int,
        engine: str = "dummy",
        input: Optional[str] = None,  # pylint: disable=redefined-builtin
    ) -> MetisRequestIdDTO:
        "Create calculation"
        async with await self._client.request(
            method="POST",
            url=self._base_url,
            json={"dataId": data_id, "engine": engine, "input": input},
            auth_required=True,
        ) as resp:
            return await resp.json()

    async def create(
        self,
        data_id: int,
        engine: str = "dummy",
        input: Optional[str] = None,  # pylint: disable=redefined-builtin
    ) -> Optional[MetisCalculationDTO]:
        "Create calculation and wait for result"
        evt = await act_and_get_result_from_stream(
            self._root.stream.subscribe,
            partial(self.create_event, data_id, engine, input),
        )
        if evt["type"] == "calculations":
            data = sorted(
                evt.get("data", {}).get("data", []),
                key=lambda x: x.get("createdAt", datetime.fromordinal(1)),
            )
            return data[-1] if data else None

    async def _get_results(
        self,
        calc_getter: Callable[[], Awaitable[Optional[MetisCalculationDTO]]],
        on_progress: Optional[MetisCalculationOnProgressT] = None,
    ) -> Optional[Sequence[MetisDataSourceDTO]]:
        "Waits for the end of the calculation and returns the results"

        def get_new_calc_id(data_id: int, calcs: Sequence[MetisCalculationDTO]):
            """
            The funny part is that when the computation ends,
            we get a data source instead of a calculation.
            """
            for calc in calcs:
                if calc.get("progress", 0) < 100:
                    continue
                data = cast(MetisDataSourceDTO, calc)
                if data_id in data.get("parents", []):
                    return data["id"]
            return None

        def get_calc_from_listing(calc_id: int, calcs: Sequence[MetisCalculationDTO]):
            for calc in calcs:
                if calc_id == calc["id"]:
                    return calc
            return None

        def filter_ds_for_calc(data_id: int, dss: Sequence[MetisDataSourceDTO]):
            return [
                ds
                for ds in dss
                if ds["type"] in DATA_SOURCE_CALC_RESULT_TYPES
                and data_id in ds["parents"]
            ]

        async with self._root.stream.subscribe() as sub:
            target_calc = await calc_getter()
            if not target_calc:
                return  # pragma: no cover
            calc_id = target_calc["id"]
            data_id = target_calc["parent"]
            async for msg in sub:
                if msg["type"] == "calculations":
                    calc_id = get_new_calc_id(data_id, msg["data"]["data"]) or calc_id
                    target_calc = get_calc_from_listing(calc_id, msg["data"]["data"])

                    # run callback if any and exit if needed
                    if target_calc and on_progress:
                        if (
                            await on_progress(target_calc)
                            if iscoroutinefunction(on_progress)
                            else on_progress(target_calc)
                        ) is False:
                            return

                # results
                if msg["type"] == "datasources":
                    results = filter_ds_for_calc(data_id, msg["data"]["data"])
                    # if results or calc is done but no results
                    if results or target_calc is None:
                        return results

    async def get_results(
        self, calc_id: int, on_progress: Optional[MetisCalculationOnProgressT] = None
    ) -> Optional[Sequence[MetisDataSourceDTO]]:
        "Waits for the end of the calculation and returns the results"

        async def get_calc():
            return await self.get(calc_id)

        return await self._get_results(get_calc, on_progress)

    async def create_get_results(
        self,
        data_id: int,
        engine: str = "dummy",
        input: Optional[str] = None,  # pylint: disable=redefined-builtin
        on_progress: Optional[MetisCalculationOnProgressT] = None,
    ) -> Optional[Sequence[MetisDataSourceDTO]]:
        "Create calculation, wait done and get results"

        async def get_calc():
            return await self.create(data_id, engine, input)

        return await self._get_results(get_calc, on_progress)

    @raise_on_metis_error
    async def get_engines(self) -> Sequence[str]:
        "Get supported calculation engines"
        async with await self._client.request(
            method="GET",
            url=self._base_url / "engines",
            auth_required=False,
        ) as resp:
            return await resp.json()

    async def list_event(self) -> MetisRequestIdDTO:
        "List all user's calculations"
        async with await self._client.request(
            method="GET",
            url=self._base_url,
            auth_required=True,
        ) as resp:
            return await resp.json()

    async def list(self) -> Sequence[MetisCalculationDTO]:
        "List all user's calculations and wait for result"
        evt = await act_and_get_result_from_stream(
            self._root.stream.subscribe, partial(self.list_event)
        )
        if evt["type"] == "calculations":
            return evt.get("data", {}).get("data", [])
        return []  # pragma: no cover

    async def get(self, calc_id: int) -> Optional[MetisCalculationDTO]:
        "Get calculation by id"
        data = list(filter(lambda x: x["id"] == calc_id, await self.list()))
        return data[-1] if data else None



"Test MetisCollectionsNamespace"

import asyncio
import json
from contextlib import nullcontext as does_not_raise
from copy import deepcopy
from datetime import datetime
from functools import partial
from typing import List

import pytest
from aiohttp import web
from aiohttp.test_utils import TestClient, TestServer
from aiohttp.web_exceptions import HTTPNotFound, HTTPOk
from yarl import URL

from metis_client import MetisAPI, MetisAPIAsync, MetisTokenAuth
from metis_client.dtos import (
    MetisCollectionCreateDTO,
    MetisCollectionDTO,
    MetisErrorDTO,
    MetisRequestIdDTO,
)
from metis_client.exc import (
    MetisNotFoundException,
    MetisPayloadException,
    MetisQuotaException,
)
from metis_client.models import MetisMessageEvent
from tests.helpers import random_word

dt = datetime.fromordinal(1)
TOKEN = random_word(10)
PATH_STREAM = "/stream"
PATH_PING = "/v0"
PATH_C = "/v0/collections"
PATH_C_ID = "/v0/collections/{id}"
PATH_C_PUT_PAYLOAD: MetisCollectionCreateDTO = {
    "title": random_word(10),
    "typeId": 1,
    "dataSources": [1, 2, 3],
    "users": [1, 2, 3],
    "description": random_word(10),
}
PATH_C_PUT_RESPONSE_PAYLOAD: MetisCollectionDTO = {
    **PATH_C_PUT_PAYLOAD,
    "createdAt": dt,
    "updatedAt": dt,
}
PATH_C_PUT_RESPONSE_ERROR_PAYLOAD: MetisErrorDTO = {"status": 400, "error": "oops"}
PATH_C_GET_RESPONSE_PAYLOAD = deepcopy(PATH_C_PUT_RESPONSE_PAYLOAD)
PATH_C_GET_RESPONSE_ERROR_PAYLOAD: MetisErrorDTO = {"status": 402, "error": "oops"}
PATH_C_DELETE_RESPONSE_ERROR_PAYLOAD = deepcopy(PATH_C_PUT_RESPONSE_ERROR_PAYLOAD)
PATH_C_DELETE_RESPONSE_ERROR404_PAYLOAD = {
    "status": HTTPNotFound.status_code,
    "error": "not found",
}

event_stream: List[MetisMessageEvent] = []


def make_collections_event(req_id: str, datas) -> MetisMessageEvent:
    "Create collections event with content"
    evt_data_dto = {"reqId": req_id, "data": datas, "total": 1, "types": []}
    return MetisMessageEvent(
        "collections", "collections", json.dumps(evt_data_dto), "", ""
    )


def make_error_event(req_id: str, datas) -> MetisMessageEvent:
    "Create errors event with content"
    evt_data_dto = {"reqId": req_id, "data": datas}
    return MetisMessageEvent("errors", "errors", json.dumps(evt_data_dto), "", "")


async def sse_handler(_: web.Request) -> web.Response:
    "Request handler"
    body = "".join(
        [f"event: {evt.type}\ndata: {evt.data}\n\n\n" for evt in event_stream]
    )
    resp = web.Response(status=HTTPOk.status_code, body=body)
    resp.content_type = "text/event-stream"
    resp.headers["Cache-Control"] = "no-store"
    return resp


async def ping_handler(_) -> web.Response:
    "Request handler"
    evt = MetisMessageEvent("", "ping", "", "", "")
    event_stream.append(evt)
    return web.Response(status=HTTPOk.status_code)


async def collection_create_handler(request: web.Request) -> web.Response:
    "Request handler"
    body: MetisRequestIdDTO = {"reqId": random_word(10)}

    payload = await request.json()
    if payload.get("typeId") == 1:
        collection_dto = {
            **payload,
            "createdAt": dt.isoformat(),
            "updatedAt": dt.isoformat(),
        }
        evt = make_collections_event(body["reqId"], [collection_dto])
    else:
        evt = make_error_event(body["reqId"], [PATH_C_PUT_RESPONSE_ERROR_PAYLOAD])

    event_stream.append(evt)

    return web.json_response(body, status=HTTPOk.status_code)


async def list_collections_handler(request: web.Request) -> web.Response:
    "Request handler"
    auth_header = request.headers.get("Authorization", "")
    body: MetisRequestIdDTO = {"reqId": random_word(10)}

    if auth_header != f"Bearer {str(True)}":
        collection_dto = {
            **PATH_C_GET_RESPONSE_PAYLOAD,
            "createdAt": dt.isoformat(),
            "updatedAt": dt.isoformat(),
        }
        evt = make_collections_event(body["reqId"], [collection_dto])
    else:
        evt = make_error_event(body["reqId"], [PATH_C_GET_RESPONSE_ERROR_PAYLOAD])

    event_stream.append(evt)

    return web.json_response(body, status=HTTPOk.status_code)


async def delete_collection_handler(request: web.Request) -> web.Response:
    "Request handler"

    if request.match_info.get("id") not in ["1", "2"]:
        return web.Response(status=HTTPNotFound.status_code)

    body: MetisRequestIdDTO = {"reqId": random_word(10)}

    if request.match_info.get("id") == "2":
        evt = make_error_event(body["reqId"], [PATH_C_DELETE_RESPONSE_ERROR_PAYLOAD])
    else:
        evt = make_collections_event(body["reqId"], [])

    event_stream.append(evt)

    return web.json_response(body, status=HTTPOk.status_code)


async def create_app() -> web.Application:
    "Create web application"
    app = web.Application()
    app.router.add_head(PATH_PING, ping_handler)
    app.router.add_get(PATH_STREAM, sse_handler)
    app.router.add_put(PATH_C, collection_create_handler)
    app.router.add_get(PATH_C, list_collections_handler)
    app.router.add_delete(PATH_C_ID, delete_collection_handler)
    return app


@pytest.fixture
async def aiohttp_client_impl(aiohttp_client) -> TestClient:
    "Create test client"
    app = await create_app()
    return await aiohttp_client(TestServer(app))


@pytest.fixture
def base_url(aiohttp_client_impl: TestClient) -> URL:
    "Return base url"
    return aiohttp_client_impl.make_url("")


@pytest.fixture
def client(base_url: URL) -> MetisAPI:
    "Return sync client"
    return MetisAPI(base_url, auth=MetisTokenAuth(TOKEN))


@pytest.fixture
async def client_async(base_url: URL):
    "Return sync client"
    async with MetisAPIAsync(base_url, auth=MetisTokenAuth(TOKEN)) as client:
        yield client


@pytest.mark.parametrize(
    "type_id, expected, raises",
    [
        (1, PATH_C_PUT_RESPONSE_PAYLOAD, does_not_raise()),
        (2, None, pytest.raises(MetisPayloadException)),
    ],
)
async def test_create_collection(
    client: MetisAPI, client_async: MetisAPIAsync, type_id, expected, raises
):
    "Test create()"
    kwargs = {
        "type_id": type_id,
        "title": PATH_C_PUT_PAYLOAD["title"],
        "description": PATH_C_PUT_PAYLOAD["description"],
        "data_source_ids": PATH_C_PUT_PAYLOAD["dataSources"],
        "user_ids": PATH_C_PUT_PAYLOAD["users"],
    }
    with raises:
        col = await client_async.v0.collections.create(**kwargs)
        assert col == expected, "Response matches"
    with raises:
        col = await asyncio.get_event_loop().run_in_executor(
            None, partial(client.v0.collections.create, **kwargs)
        )
        assert col == expected, "Response matches"


@pytest.mark.parametrize(
    "fail, expected, raises",
    [
        (False, [PATH_C_GET_RESPONSE_PAYLOAD], does_not_raise()),
        (True, None, pytest.raises(MetisQuotaException)),
    ],
)
async def test_list_collections(base_url: URL, fail: bool, expected, raises):
    "Test list()"
    with raises:
        async with MetisAPIAsync(base_url, auth=MetisTokenAuth(str(fail))) as client:
            cols = await client.v0.collections.list()
        assert cols == expected, "Response matches"
    with raises:
        client = MetisAPI(base_url, auth=MetisTokenAuth(str(fail)))
        cols = await asyncio.get_event_loop().run_in_executor(
            None, client.v0.collections.list
        )
        assert cols == expected, "Response matches"


@pytest.mark.parametrize(
    "col_id, raises",
    [
        (1, does_not_raise()),
        (2, pytest.raises(MetisPayloadException)),
        (3, pytest.raises(MetisNotFoundException)),
    ],
)
async def test_delete_collection(
    client: MetisAPI, client_async: MetisAPIAsync, col_id: int, raises
):
    "Test delete()"
    with raises:
        await client_async.v0.collections.delete(col_id)
    with raises:
        await asyncio.get_event_loop().run_in_executor(
            None, client.v0.collections.delete, col_id
        )



"Test MetisSubscription"

import logging

from metis_client.dtos.event import MetisErrorEventDTO
from metis_client.models import MetisHub, MetisSubscription

TEST_EVENT: MetisErrorEventDTO = {
    "type": "errors",
    "data": {"reqId": "1", "data": [{"status": 1, "error": "oops"}]},
}


async def test_queue_full(caplog):
    "Test internal queue is full"
    hub = MetisHub()
    sub = MetisSubscription(hub, queue_size=1)
    sub.put_nowait(TEST_EVENT)
    with caplog.at_level(logging.WARNING):
        sub.put_nowait(TEST_EVENT)
    assert "queue" in caplog.text
    await hub.close()


async def test_len():
    "Test subscription length"
    hub = MetisHub()
    sub = MetisSubscription(hub, queue_size=1)
    sub.put_nowait(TEST_EVENT)
    assert len(sub) == 1, "Internal queue size match"
    await hub.close()



"Test MetisError"

from metis_client.exc import MetisError


def test_error_string():
    "Test MetisError"
    status = -1234
    message = "mEsSaGe"
    err = MetisError(status=status, message=message)
    assert str(status) in str(err), "Status should be printed"
    assert str(message) in str(err), "Message should be printed"



"Test MetisNoAuth"
import pytest
from aiohttp import web
from aiohttp.test_utils import TestClient, TestServer

from metis_client import MetisNoAuth


@pytest.fixture
async def cli(aiohttp_client) -> TestClient:
    "Create test client"
    return await aiohttp_client(TestServer(web.Application()))


async def test_auth(cli: TestClient):
    "Test no authentication"
    authenticator = MetisNoAuth()
    base_url = cli.make_url("")
    assert (
        await authenticator.should_update(cli.session, base_url) is False
    ), "Update is not needed before authentication"
    assert await authenticator.authenticate(
        cli.session, base_url
    ), "Authenticaion is successful"
    assert (
        await authenticator.should_update(cli.session, base_url) is False
    ), "Update is not needed after authentication"



"Some test helpers"
import random
import string


def random_word(length):
    "Generate random string of length"
    letters = string.ascii_lowercase
    return "".join(random.choice(letters) for _ in range(length))
